/* $NetBSD$ */

/*-
 * Copyright (c) 2009 Jared D. McNeill <jmcneill@invisible.ca>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/dkio.h>
#include <sys/disklabel.h>
#include <sys/stat.h>

#include <dev/pud/pud_msgif.h>
#include <dev/pud/pud_msgif.h>

#include <prop/proplib.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libvdisk.h"

#define	VDISK_ARBITRARY_UNIT_LIMIT	256
#define	VDISK_PUD_REQBUF_LEN		(1024 * 1024)
#define VDISK_DEBUG_DEFAULT_IMPORTANCE	"30" /* messages 30% or higher get displayed */

#ifndef USE_ARG
#define USE_ARG(x)  /*LINTED*/(void)&x
#endif

static bool vdisk_dying = false;

static bool
vdisk_free_device(struct vdisk *vd, int unit)
{
	USE_ARG(vd);
	USE_ARG(unit);
	vdisk_report(30, "vdisk_free_device todo\n");
	return false;
}

/* create the device nodes - all the partitions on the device */
static bool
vdisk_alloc_device_nodes(struct vdisk *vd, unsigned mode, int unit)
{
	char devpath[PATH_MAX];
	dev_t dev;
	int part;

	USE_ARG(vd);
	for (part = 0; part < VDISK_NUM_PARTITIONS; part++) {
		dev = MAKEDISKDEV(VDISK_MAJOR, unit, part);
		(void) snprintf(devpath, sizeof(devpath),
		    "/dev/vdisk/vd%d%c", unit, part + 'a');
		vdisk_report(20, "creating vd%d%c and rvd%d%c (%d,%d)\n",
		    unit, part + 'a', unit, part + 'a',
		    major(dev), minor(dev));
		if (mknod(devpath, mode | S_IFBLK, dev) == -1) {
			vdisk_report(50, "mknod(%s, %o, %x) failed: %s\n",
			    devpath, mode, (unsigned)dev, strerror(errno));
			return false;
		}
		(void) snprintf(devpath, sizeof(devpath),
		    "/dev/vdisk/rvd%d%c", unit, part + 'a');
		if (mknod(devpath, mode | S_IFCHR, dev) == -1) {
			vdisk_report(50, "mknod(%s, %o, %x) failed: %s\n",
			    devpath, mode, (unsigned)dev, strerror(errno));
			return false;
		}
	}

	return true;
}

/* XXX synchronization */
static bool
vdisk_alloc_device(struct vdisk *vd, unsigned mode)
{
	struct stat st;
	int unit, part;

	if (stat("/dev/vdisk", &st) == -1) {
		if (mkdir("/dev/vdisk", 0755) == -1) {
			vdisk_report(50, "couldn't create /dev/vdisk: %s",
			    strerror(errno));
			return false;
		}
	}
	for (unit = 0; unit < VDISK_ARBITRARY_UNIT_LIMIT; unit++) {
		if (!vdisk_alloc_device_nodes(vd, mode, unit)) {
			vdisk_free_device(vd, unit);
			return false;
		}
		vd->vd_unit = unit;
		for (part = 0; part < VDISK_NUM_PARTITIONS; part++) {
			vd->vd_vp[part].vp_regdev =
			    MAKEDISKDEV(VDISK_MAJOR, unit, part);
			vd->vd_vp[part].vp_part = part;
			vd->vd_vp[part].vp_vd = vd;
		}
		break;
		/*NOTREACHED*/
	}
	return true;
}

static bool
vdisk_open_pud(struct vdisk *vd)
{
	int part;

	for (part = 0; part < VDISK_NUM_PARTITIONS; part++) {
		vd->vd_vp[part].vp_fd = open(_PATH_PUD, O_RDWR);
		if (vd->vd_vp[part].vp_fd == -1) {
			vdisk_report(20, "open %s: %s\n", _PATH_PUD, strerror(errno));
			return false;
		}
	}

	return true;
}

static void
vdisk_close_pud(struct vdisk *vd)
{
	int part;

	if (vd == NULL)
		return;

	for (part = 0; part < VDISK_NUM_PARTITIONS; part++) {
		if (vd->vd_vp[part].vp_fd == -1)
			continue;
		close(vd->vd_vp[part].vp_fd);
		vd->vd_vp[part].vp_fd = -1;
	}
}

static struct vdisk_part *
vdisk_fd_to_part(struct vdisk *vd, int fd)
{
	int part;

	for (part = 0; part < VDISK_NUM_PARTITIONS; part++)
		if (vd->vd_vp[part].vp_fd == fd)
			return &vd->vd_vp[part];

	return NULL;
}

static int
vdisk_do_pud(struct vdisk_part *vp)
{
	struct vdisk *vd = vp->vp_vd;
	struct vdisk_if *vif = &vd->vd_if;
	struct pud_req *pdr = vd->vd_pdr;
	struct pud_creq_open *pm_open;
	struct pud_creq_close *pm_close;
	struct pud_creq_read *pm_read;
	struct pud_creq_write *pm_write;
	struct pud_req_ioctl *pm_ioctl;
	size_t len;
	long n;

	n = read(vp->vp_fd, vd->vd_pdr, vd->vd_pdrlen);
	if (n == -1) {
		vdisk_report(50, "read pud request failed: %s\n", strerror(errno));
		return -1;
	}

	switch (pdr->pdr_reqtype) {
	case PUD_CDEV_OPEN:
		pm_open = (struct pud_creq_open *)(void *)pdr;
		pdr->pdr_rv = 0;
		if (vif->open)
			pdr->pdr_rv = vif->open(vd->vd_cookie,
			    pm_open->pm_flags, pm_open->pm_fmt);
		break;
	case PUD_CDEV_CLOSE:
		pm_close = (struct pud_creq_close *)(void *)pdr;
		pdr->pdr_rv = 0;
		if (vif->close)
			pdr->pdr_rv = vif->close(vd->vd_cookie,
			    pm_close->pm_flags, pm_close->pm_fmt);
		break;
	case PUD_CDEV_READ:
		pm_read = (struct pud_creq_read *)(void *)pdr;
		vdisk_report(20, "read from offset %llu, resid %zu\n",
		    (unsigned long long)pm_read->pm_offset,
		    pm_read->pm_resid);
		len = pm_read->pm_resid;
		vif->read(vd->vd_cookie, pm_read->pm_data,
		    pm_read->pm_offset, &len);
		if (pdr->pdr_reqclass == PUD_REQ_BDEV) {
			len = pm_read->pm_resid;
			pm_read->pm_resid = 0;
		} else
			pm_read->pm_resid -= len;

		pdr->pdr_pth.pth_framelen = sizeof(*pm_read) + len;

		/* XXX */
		pdr->pdr_rv = 0;
		break;
	case PUD_CDEV_WRITE:
		pm_write = (struct pud_creq_write *)(void *)pdr;
		vdisk_report(20, "write to offset %llu, resid %zu\n",
		    (unsigned long long)pm_write->pm_offset,
		    pm_write->pm_resid);
		vif->write(vd->vd_cookie, pm_write->pm_data,
		    pm_write->pm_offset, &pm_write->pm_resid);

		pdr->pdr_pth.pth_framelen = sizeof(*pm_write);

		/* XXX */
		pdr->pdr_rv = 0;
		break;

	case PUD_CDEV_IOCTL:
		pm_ioctl = (struct pud_req_ioctl *)(void *)pdr;
		pdr->pdr_rv = vdisk_ioctl(vp,
		    pm_ioctl->pm_iocmd, pm_ioctl->pm_data, pm_ioctl->pm_flag);
		break;
	default:
		vdisk_report(50, "unhandled request type %d\n", pdr->pdr_reqtype);
		pdr->pdr_rv = EINVAL;
		break;
	}

	n = write(vp->vp_fd, pdr, pdr->pdr_pth.pth_framelen);
	if (n == -1) {
		vdisk_report(50, "write pud response failed: %s\n", strerror(errno));
		return -1;
	}

	return 0;
}

/* print message if we want it */
int
vdisk_report(const int importance, const char *fmt, ...)
{
	const char	*envlevel;
	va_list		 args;

	if ((envlevel = getenv("VDISK_DEBUG")) == NULL) {
		envlevel = VDISK_DEBUG_DEFAULT_IMPORTANCE;
	}
	if (importance >= atoi(envlevel)) {
		(void) fprintf(stderr, "libvdisk: ");
		va_start(args, fmt);
		vfprintf(stderr, fmt, args);
		va_end(args);
		return 1;
	}
	return 0;
}

struct vdisk *
vdisk_init(const char *name, int mode, void *cookie)
{
	struct vdisk *vd;
	struct pud_conf_reg pcr;
	int part;
	long n;

	USE_ARG(name);
	vd = calloc(1, sizeof(*vd));
	if (vd == NULL)
		return NULL;

	for (part = 0; part < VDISK_NUM_PARTITIONS; part++)
		vd->vd_vp[part].vp_fd = -1;
	if (!vdisk_open_pud(vd)) {
		vdisk_close_pud(vd);
		free(vd);
		return NULL;
	}

	vd->vd_cookie = cookie;
	vd->vd_pdr = malloc(VDISK_PUD_REQBUF_LEN);
	if (vd->vd_pdr == NULL) {
		vdisk_close_pud(vd);
		free(vd);
		return NULL;
	}
	vd->vd_pdrlen = VDISK_PUD_REQBUF_LEN;
	if (!vdisk_alloc_device(vd, (unsigned)mode)) {
		vdisk_close_pud(vd);
		free(vd->vd_pdr);
		free(vd);
		return NULL;
	}

	/* register device numbers */
	memset(&pcr, 0, sizeof(pcr));
	pcr.pm_pdr.pdr_pth.pth_framelen = sizeof(pcr);
	pcr.pm_version = (int)(PUD_DEVELVERSION | PUD_VERSION);
	pcr.pm_pdr.pdr_reqclass = PUD_REQ_CONF;
	pcr.pm_pdr.pdr_reqtype = PUD_CONF_REG;

	for (part = 0; part < VDISK_NUM_PARTITIONS; part++) {
		(void) snprintf(pcr.pm_devname, sizeof(pcr.pm_devname),
		    "vdisk/vd%d%c", vd->vd_unit, part + 'a');
		pcr.pm_regdev = vd->vd_vp[part].vp_regdev;
		pcr.pm_flags = PUD_CONFFLAG_BDEV | PUD_CONFFLAG_DISK;

		vdisk_report(20, "registering %s (%d,%d/%d)\n", pcr.pm_devname,
		    major(pcr.pm_regdev), minor(pcr.pm_regdev),
		    (unsigned)pcr.pm_regdev);
		n = write(vd->vd_vp[part].vp_fd, &pcr, sizeof(pcr));
		if (n == -1 || pcr.pm_pdr.pdr_rv != 0) {
			vdisk_close_pud(vd);
			free(vd->vd_pdr);
			free(vd);
			return NULL;
		}
	}

	vdisk_report(20, "vdisk created\n");
	return vd;
}

void
vdisk_main(struct vdisk *vd)
{
	const struct vdisk_if *vif = &vd->vd_if;
	struct vdisk_part *vp;
	struct timeval timeo;
	struct fd_set fds;
	int part, i, n, maxfd = -1;

	assert(vif->read != NULL);
	assert(vif->write != NULL);
	assert(vif->sizedev != NULL);

	vdisk_getdisklabel(vd);

	while (!vdisk_dying) {
		FD_ZERO(&fds);
		for (part = 0; part < VDISK_NUM_PARTITIONS; part++) {
			FD_SET(vd->vd_vp[part].vp_fd, &fds);
			if (vd->vd_vp[part].vp_fd > maxfd)
				maxfd = vd->vd_vp[part].vp_fd;
		}
		timeo.tv_sec = 1;
		timeo.tv_usec = 0;
		n = select(maxfd + 1, &fds, NULL, NULL, &timeo);
		switch (n) {
		case -1:
			if (errno != EINTR)
				vdisk_report(50, "select failed: %s\n", strerror(errno));
			return;
		case 0:
			break;
		default:
			vdisk_report(20, "pud request on %d partition(s)\n", n);
			for (i = 0; i <= maxfd; i++)
				if (FD_ISSET(i, &fds)) {
					vp = vdisk_fd_to_part(vd, i);
					if (vp)
						vdisk_do_pud(vp);
				}
			break;
		}
	}
}

void
vdisk_fini(struct vdisk *vd)
{
	char devpath[PATH_MAX];
	int part;

	if (vd == NULL)
		return;

	for (part = 0; part < VDISK_NUM_PARTITIONS; part++) {
		snprintf(devpath, sizeof(devpath),
		    "/dev/vdisk/vd%d%c", vd->vd_unit, part + 'a');
		if (unlink(devpath) == -1) {
			vdisk_report(50, "couldn't remove %s: %s\n", devpath,
			    strerror(errno));
		} else {
			vdisk_report(20, "removed %s\n", devpath);
		}

		snprintf(devpath, sizeof(devpath),
		    "/dev/vdisk/rvd%d%c", vd->vd_unit, part + 'a');
		if (unlink(devpath) == -1) {
			vdisk_report(50, "couldn't remove %s: %s\n", devpath,
			    strerror(errno));
		} else {
			vdisk_report(20, "removed %s\n", devpath);
		}
	}
}

void
vdisk_signal(int sig)
{
	USE_ARG(sig);
	vdisk_dying = true;
}
