/* $NetBSD$ */

/*-
 * Copyright (c) 2009 Jared D. McNeill <jmcneill@invisible.ca>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/dkio.h>
#include <sys/disklabel.h>
#include <sys/stat.h>
#include <sys/param.h>

#include <dev/pud/pud_msgif.h>
#include <dev/pud/pud_disk.h>

#include <prop/proplib.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "vdisk.h"

#ifndef USE_ARG
#define USE_ARG(x)  /*LINTED*/(void)&x
#endif

#define VDISK_DUMMY_IMGSIZE	(64 * 1024 * 1024)

struct vdisk_dummy_priv {
	uint64_t	 size;
	uint8_t		*buf;
};

static void
vdisk_dummy_read(void *opaque, uint8_t *data, off_t offset, size_t *resid)
{
	struct vdisk_dummy_priv *priv = opaque;

	if (*resid == 0 || offset + *resid > priv->size) {
		vdisk_report(30, "vdisk_dummy_read ignored\n");
		return;
	}

	vdisk_report(20, "vdisk_dummy_read offset %llu resid %zu\n",
	    (unsigned long long)offset, *resid);
	memcpy(data, priv->buf + offset, *resid);
}

static void
vdisk_dummy_write(void *opaque, uint8_t *data, off_t offset, size_t *resid)
{
	struct vdisk_dummy_priv *priv = opaque;

	if (*resid == 0 || offset + *resid > priv->size) {
		vdisk_report(30, "vdisk_dummy_write ignored\n");
		return;
	}

	vdisk_report(20, "vdisk_dummy_write offset %llu resid %zu\n",
	    (unsigned long long)offset, *resid);
	memcpy(priv->buf + offset, data, *resid);

	*resid = 0;
}

static void
vdisk_dummy_sizedev(void *opaque, uint64_t *devsz)
{
	struct vdisk_dummy_priv *priv = opaque;

	*devsz = priv->size;
}

int
main(int argc, char **argv)
{
	struct vdisk_dummy_priv	 priv;
	struct vdisk		*vd;
	int			 i;

	(void) memset(&priv, 0x0, sizeof(priv));
	priv.size = VDISK_DUMMY_IMGSIZE;
	while ((i = getopt(argc, argv, "s:")) != -1) {
		switch(i) {
		case 's':
			priv.size = atoi(optarg);
			break;
		default:
			break;
		}
	}
	vd = vdisk_init("debug", 0640, NULL);
	if (vd == NULL)
		printf("failed to initialise vdisk\n");
	else {
		priv.buf = calloc(1, priv.size);
		if (priv.buf == NULL) {
			perror("malloc");
			goto cleanup;
		}

		vd->vd_cookie = &priv;
		vd->vd_if.read = vdisk_dummy_read;
		vd->vd_if.write = vdisk_dummy_write;
		vd->vd_if.sizedev = vdisk_dummy_sizedev;

		printf("/dev/vdisk/vd%d%c\n", vd->vd_unit, RAW_PART + 'a');
		signal(SIGINT, vdisk_signal);
		vdisk_main(vd);
	}

cleanup:
	vdisk_fini(vd);

	return 0;
}
