/*-
 * Copyright (c) 2010 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uuid.h>

#include <sys/endian.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "vhd.h"

#define HEADER_CHECKSUM_OFFSET	36
#define FOOTER_CHECKSUM_OFFSET	64

/* structure to map a string to a numeric value */
typedef struct strmap_t {
	const char	*s;	/* string */
	const int	 n;	/* numeric value */
} strmap_t;

/* given a numeric value, find its string representation */
static const char *
findstr(strmap_t *mappings, uint32_t n)
{
	strmap_t	*map;

	for (map = mappings ; map->s && (int)n != map->n ; map++) {
	}
	return map->s;
}

/* calculate the (simple) checksum for an area */
static uint32_t
cksum(void *footer, size_t size, uint32_t ig)
{
	uint32_t	 ret;
	uint8_t		*s;
	size_t		 i;

	s = (uint8_t *)footer;
	for (ret = 0, i = 0 ; i < size ; i++) {
		if (i == ig) {
			i += sizeof(ig) - 1;
		} else {
			ret += s[i];
		}
	}
	return ~ret;
}

/* table of VHD disk types */
static strmap_t	disktypes[] = {
	{	"None",		VHD_TYPE_NONE		},
	{	"Reserved",	VHD_TYPE_RESERVED	},
	{	"Fixed",	VHD_TYPE_FIXED		},
	{	"Dynamic",	VHD_TYPE_DYNAMIC	},
	{	"Differencing",	VHD_TYPE_DIFFERENCING	},
	{	"Reserved5",	VHD_TYPE_RESERVED5	},
	{	"Reserved6",	VHD_TYPE_RESERVED6	},
	{	NULL,		-1			},
};

/* populate a footer structure */
static void
fillfooter(vhd_footer_t *footer, uint32_t disktype, uint64_t size, uint32_t geom)
{
	(void) memset(footer, 0x0, sizeof(*footer));
	(void) snprintf((char *)footer->cookie, sizeof(footer->cookie), "NetBSD");
	footer->version = VHD_FORMAT_VERSION;
	footer->data_offset = VHD_FIXED_OFFSET;
	footer->timestamp = SECS_TO_VHD_TIME(time(NULL));
	(void) memcpy(footer->maker_app, "vhd ", sizeof(footer->maker_app));
	footer->maker_version = NETBSD_VHD;
	(void) memcpy(&footer->maker_host_os, VHD_HOST_OS_NETBSD,
			sizeof(footer->maker_host_os));
	footer->origsize = footer->size = size;
	footer->geom = geom;
	footer->disktype = disktype;
	footer->cksum = cksum(footer, sizeof(*footer), FOOTER_CHECKSUM_OFFSET);
	uuidgen((struct uuid *)(void *)footer->uuid, 1);
}

/* byte swap a footer structure in-place */
static void
swapfooter(vhd_footer_t *footer, const char *direction)
{
	if (strcmp(direction, "hton") == 0) {
		footer->version = htobe32(footer->version);
		footer->data_offset = htobe64(footer->data_offset);
		footer->timestamp = htobe32(footer->timestamp);
		footer->maker_version = htobe32(footer->maker_version);
		footer->origsize = htobe64(footer->origsize);
		footer->size = htobe64(footer->size);
		footer->geom = htobe32(footer->geom);
		footer->disktype = htobe32(footer->disktype);
		footer->cksum = htobe32(footer->cksum);
	} else {
		footer->version = be32toh(footer->version);
		footer->data_offset = be64toh(footer->data_offset);
		footer->timestamp = be32toh(footer->timestamp);
		footer->maker_version = be32toh(footer->maker_version);
		footer->origsize = be64toh(footer->origsize);
		footer->size = be64toh(footer->size);
		footer->geom = be32toh(footer->geom);
		footer->disktype = be32toh(footer->disktype);
		footer->cksum = be32toh(footer->cksum);
	}
}

/* populate a header structure */
static void
fillheader(vhd_header_t *header, uint64_t size, uint32_t blocksize)
{
	(void) memcpy(header->cookie, VHD_HEADER_COOKIE, sizeof(header->cookie));
	header->data_offset = VHD_FIXED_OFFSET;
	header->version = VHD_FORMAT_VERSION;
	header->table_offset = sizeof(*header) + sizeof(vhd_footer_t);
	header->maxent = size / blocksize;
	header->blocksize = blocksize;
	header->cksum = cksum(header, sizeof(*header), HEADER_CHECKSUM_OFFSET);
}

/* byte swap a  header structure in-place */
static void
swapheader(vhd_header_t *header, const char *direction)
{
	if (strcmp(direction, "hton") == 0) {
		header->data_offset = htobe64(header->data_offset);
		header->version = htobe32(header->version);
		header->table_offset = htobe64(header->table_offset);
		header->maxent = htobe32(header->maxent);
		header->blocksize = htobe32(header->blocksize);
		header->cksum = htobe32(header->cksum);
	} else {
		header->data_offset = be64toh(header->data_offset);
		header->version = be32toh(header->version);
		header->table_offset = be64toh(header->table_offset);
		header->maxent = be32toh(header->maxent);
		header->blocksize = be32toh(header->blocksize);
		header->cksum = be32toh(header->cksum);
	}
}

/* print a footer structure */
static void
pfooter(vhd_footer_t *footer)
{
	uint32_t	calc;
	time_t		t;
	size_t		i;

	printf("Footer Cookie: %.8s\n", footer->cookie);
	printf("Disk features: 0x%x\n", footer->features);
	printf("Disk format version: 0x%x\n", footer->version);
	printf("Disk data offset: %" PRIu64 "\n", footer->data_offset);
	t = VHD_TIME_TO_SECS(footer->timestamp);
	printf("Disk timestamp: %.24s\n", ctime(&t));
	printf("Disk maker application: %.4s\n", footer->maker_app);
	printf("Disk maker version: 0x%x\n", footer->maker_version);
	printf("Disk maker application: %.4s\n", footer->maker_host_os);
	printf("Disk original size: %" PRIu64 "\n", footer->origsize);
	printf("Disk size: %" PRIu64 "\n", footer->size);
	printf("Disk cylinders: %d\n", VDISK_CYLS(footer->geom));
	printf("Disk heads: %d\n", VDISK_HEADS(footer->geom));
	printf("Disk sectors: %d\n", VDISK_SECTORS(footer->geom));
	printf("Disk disktype: %d %s\n", footer->disktype,
		findstr(disktypes, footer->disktype));
	calc = cksum(footer, sizeof(*footer), FOOTER_CHECKSUM_OFFSET);
	printf("Disk checksum: 0x%x,%s calculated 0x%x\n", footer->cksum,
		(calc == footer->cksum) ? "" : " [MISMATCH]", calc);
	printf("Disk uuid: ");
	for (i = 0 ; i < sizeof(footer->uuid) ; i++) {
		printf("%02x", footer->uuid[i]);
	}
	printf("\n");
	printf("Disk saved state: %d\n\n", footer->saved_state);
}

/* print a header struct */
static void
pheader(vhd_header_t *header)
{
	uint32_t	calc;

	printf("Header Cookie: %.8s\n", header->cookie);
	printf("Header data offset: %" PRIu64 "\n", header->data_offset);
	printf("Header BAT offset: %" PRIu64 "\n", header->table_offset);
	printf("Header maximum table entries: %d\n", header->maxent);
	printf("Header format version: 0x%x\n", header->version);
	printf("Header blocksize: %d\n", header->blocksize);
	calc = cksum(header, sizeof(*header), HEADER_CHECKSUM_OFFSET);
	printf("Header checksum: 0x%x,%s calculated 0x%x\n", header->cksum,
		(calc == header->cksum) ? "" : " [MISMATCH]", calc);
}

/* fill the bat table */
static uint32_t
fillbat(uint32_t **bat, uint64_t size, uint32_t blocksize, uint32_t off)
{
	uint32_t	maxent;
	uint32_t	i;

	maxent = size / blocksize;
	if ((*bat = calloc(sizeof(**bat), maxent)) == NULL) {
		(void) fprintf(stderr, "fillbat: can't allocate %u slots in bat\n", maxent);
		return 0;
	}
	for (i = 0 ; i < maxent ; i++) {
		(*bat)[i] = off + (i * blocksize);
	}
	return maxent;
}

/* byte swap a BAT table in-place */
static void
swapbat(uint32_t *bat, uint32_t maxent, const char *direction)
{
	uint32_t	i;

	if (strcmp(direction, "hton") == 0) {
		for (i = 0 ; i < maxent ; i++) {
			bat[i] = htobe32(bat[i]);
		}
	} else {
		for (i = 0 ; i < maxent ; i++) {
			bat[i] = be32toh(bat[i]);
		}
	}
}

/* print the bat table */
static void
pbat(uint32_t *bat, uint32_t maxent)
{
	uint32_t	i;

	printf("\nBAT table:\n");
	for (i = 0 ; i < maxent ; i++) {
		printf("BAT[%u]\t%u\n", i, bat[i]);
	}
}

/******************** exported functions *****************************/

/* create a vhd header, footer and BAT */
int
vhd_create(uint8_t *image, const char *vhdtype, uint64_t size, uint32_t blocksize, uint32_t geom)
{
	vhd_footer_t	 footer;
	vhd_header_t	 header;
	uint32_t	 maxent;
	uint32_t	*bat;
	uint64_t	 off;

	if (strcasecmp(vhdtype, "fixed") == 0) {
		(void) memset(&footer, 0x0, sizeof(footer));
		fillfooter(&footer, VHD_TYPE_FIXED, size, geom);
		swapfooter(&footer, "hton");
		footer.cksum = cksum(&footer, sizeof(footer), FOOTER_CHECKSUM_OFFSET);
		footer.cksum = htobe32(footer.cksum);
		(void) memcpy(&image[size], &footer, sizeof(footer));
		return 1;
	}
	if (strcasecmp(vhdtype, "dynamic") == 0) {
		/* first write footer */
		(void) memset(&footer, 0x0, sizeof(footer));
		fillfooter(&footer, VHD_TYPE_DYNAMIC, size, geom);
		swapfooter(&footer, "hton");
		footer.cksum = cksum(&footer, sizeof(footer), FOOTER_CHECKSUM_OFFSET);
		footer.cksum = htobe32(footer.cksum);
		(void) memcpy(image, &footer, sizeof(footer));
		off = sizeof(footer);
		/* then the header */
		(void) memset(&header, 0x0, sizeof(header));
		fillheader(&header, size, blocksize);
		maxent = header.maxent = fillbat(&bat, size, blocksize, off + sizeof(header));
		swapheader(&header, "hton");
		(void) memcpy(&image[off], &header, sizeof(header));
		off += sizeof(header);
		/* now the BAT table */
		swapbat(bat, maxent, "hton");
		(void) memcpy(&image[off], bat, maxent * sizeof(*bat));
		off += maxent * sizeof(*bat);
		/* then the footer again */
		(void) memcpy(&image[off], &footer, sizeof(footer));
	}
	return 1;
}

/* read a vhd header, footer and BAT table */
int
vhd_list(const char *f)
{
	vhd_footer_t	 footer;
	vhd_header_t	 header;
	struct stat	 st;
	uint32_t	 batsize;
	uint32_t	*bat;
	FILE		*fp;

	(void) memset(&footer, 0x0, sizeof(footer));
	(void) memset(&header, 0x0, sizeof(header));
	if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "listvhd: can't open '%s'\n", f);
		return 0;
	}
	(void) fstat(fileno(fp), &st);
	if (pread(fileno(fp), &footer, sizeof(footer), st.st_size - sizeof(footer)) != sizeof(footer)) {
		(void) fprintf(stderr, "listvhd: bad read '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	printf("[%s]\n", f);
	swapfooter(&footer, "ntoh");
	switch (footer.disktype) {
	case VHD_TYPE_FIXED:
		pfooter(&footer);
		break;
	case VHD_TYPE_DYNAMIC:
		if (pread(fileno(fp), &footer, sizeof(footer), 0) != sizeof(footer) ||
		    pread(fileno(fp), &header, sizeof(header), sizeof(footer)) != sizeof(header)) {
			(void) fprintf(stderr, "listvhd: bad read '%s'\n", f);
			(void) fclose(fp);
			return 0;
		}
		swapfooter(&footer, "ntoh");
		pfooter(&footer);
		swapheader(&header, "ntoh");
		pheader(&header);
		batsize = header.maxent * sizeof(*bat);
		bat = calloc(sizeof(*bat), header.maxent);
		if (pread(fileno(fp), bat, batsize, sizeof(footer) + sizeof(header)) != batsize) {
			(void) fprintf(stderr, "listvhd: bad bat read '%s'\n", f);
			(void) fclose(fp);
			return 0;
		}
		swapbat(bat, header.maxent, "ntoh");
		pbat(bat, header.maxent);
		free(bat);
		break;
	default:
		(void) fprintf(stderr, "listvhd: no such disk type '%d'\n", footer.disktype);
		break;
	}
	(void) fclose(fp);
	return 1;
}

/* read the footer from the vhd disk */
int
vhd_read_footer(FILE *fp, vhd_footer_t *footer)
{
	struct stat	st;

	(void) fstat(fileno(fp), &st);
	if (pread(fileno(fp), footer, sizeof(*footer), st.st_size - sizeof(*footer)) != sizeof(*footer)) {
		return 0;
	}
	swapfooter(footer, "ntoh");
	return 1;
}

/* read a vhd header, footer and BAT table */
int
vhd_attach(const char *f, FILE *fp, vhd_header_t *header, vhd_footer_t *footer, uint32_t **bat)
{
	uint32_t	batsize;

	(void) memset(footer, 0x0, sizeof(*footer));
	(void) memset(header, 0x0, sizeof(*header));
	if (!vhd_read_footer(fp, footer)) {
		(void) fprintf(stderr, "listvhd: bad read '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	printf("[%s]\n", f);
	switch (footer->disktype) {
	case VHD_TYPE_FIXED:
		break;
	case VHD_TYPE_DYNAMIC:
		if (pread(fileno(fp), footer, sizeof(*footer), 0) != sizeof(*footer) ||
		    pread(fileno(fp), header, sizeof(*header), sizeof(*footer)) != sizeof(*header)) {
			(void) fprintf(stderr, "listvhd: bad read '%s'\n", f);
			return 0;
		}
		swapfooter(footer, "ntoh");
		swapheader(header, "ntoh");
		batsize = header->maxent * sizeof(**bat);
		*bat = calloc(sizeof(**bat), header->maxent);
		if (pread(fileno(fp), *bat, batsize, sizeof(*footer) + sizeof(*header)) != batsize) {
			(void) fprintf(stderr, "listvhd: bad bat read '%s'\n", f);
			return 0;
		}
		swapbat(*bat, header->maxent, "ntoh");
		break;
	default:
		(void) fprintf(stderr, "listvhd: no such disk type '%d'\n", footer->disktype);
		break;
	}
	return 1;
}

/* return the size in the host file system of the vhd */
uint64_t
vhd_size(const char *vhdtype, uint64_t size, uint32_t blocksize)
{
	if (strcasecmp(vhdtype, "fixed") == 0) {
		return size + sizeof(vhd_footer_t);
	}
	if (strcasecmp(vhdtype, "dynamic") == 0) {
		return (sizeof(vhd_footer_t) * 2) + sizeof(vhd_header_t) +
			(size / blocksize) * sizeof(uint32_t);
	}
	(void) fprintf(stderr, "vhdtype '%s' not recognised\n", vhdtype);
	return 0;
}
