/*-
 * Copyright (c) 2010 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/dkio.h>
#include <sys/disklabel.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <sys/mount.h>

#include <dev/pud/pud_msgif.h>
#include <dev/pud/pud_disk.h>

#include <prop/proplib.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "libvdisk.h"
#include "vdisk.h"
#include "vhd.h"

#ifndef USE_ARG
#define USE_ARG(x)  /*LINTED*/(void)&x
#endif

struct vdisk_dummy_priv {
	uint64_t	 size;
	uint8_t		*buf;
};

/* write out the image to host disk */
static int
writeout(const char *f, const char *image, uint64_t size)
{
	uint64_t	 cc;
	FILE		*fp;
	int		 wc;

	if ((fp = fopen(f, "w")) == NULL) {
		(void) fprintf(stderr, "writeout: can't open '%s'\n", f);
		return 0;
	}
	for (cc = 0 ; cc < size && (wc = write(fileno(fp), &image[cc], size - cc)) > 0 ; cc += wc) {
	}
	(void) fclose(fp);
	return cc == size;
}

/* jared's tame implementation of a memory disk */
static void
vdisk_dummy_read(void *opaque, uint8_t *data, off_t offset, size_t *resid)
{
	struct vdisk_dummy_priv *priv = opaque;

	if (*resid == 0 || offset + *resid > priv->size) {
		vdisk_report(30, "vdisk_dummy_read ignored\n");
		return;
	}
	vdisk_report(20, "vdisk_dummy_read offset %llu resid %zu\n",
	    (unsigned long long)offset, *resid);
	memcpy(data, priv->buf + offset, *resid);
}

/* jared's tame implementation of a memory disk */
static void
vdisk_dummy_write(void *opaque, uint8_t *data, off_t offset, size_t *resid)
{
	struct vdisk_dummy_priv *priv = opaque;

	if (*resid == 0 || offset + *resid > priv->size) {
		vdisk_report(30, "vdisk_dummy_write ignored\n");
		return;
	}
	vdisk_report(20, "vdisk_dummy_write offset %llu resid %zu\n",
	    (unsigned long long)offset, *resid);
	memcpy(priv->buf + offset, data, *resid);
	*resid = 0;
}

/* jared's tame implementation of a memory disk */
static void
vdisk_dummy_sizedev(void *opaque, uint64_t *devsz)
{
	struct vdisk_dummy_priv *priv = opaque;

	*devsz = priv->size;
}

/* from kern/subr_disk.c */
static u_int
dkcksum(struct disklabel *d)
{
	size_t npartitions = d->d_npartitions;
	u_short *start, *end;
	u_short sum = 0;

	start = (u_short *)(void *)d;
	end = (u_short *)(void *)&d->d_partitions[npartitions];
	while (start < end)
		sum ^= *start++;
	return sum;
}

/* set values in the disklabel - shouldn't really be here */
static void
setlabel(struct disklabel *d, uint64_t size, uint32_t blocksize, uint32_t geom)
{
	d->d_secsize = DEV_BSIZE;
	d->d_nsectors = VDISK_SECTORS(geom);
	d->d_ntracks = VDISK_HEADS(geom);
	d->d_ncylinders = VDISK_CYLS(geom);
	d->d_secpercyl = d->d_ntracks * d->d_nsectors;
	d->d_type = DTYPE_LD;
	strlcpy(d->d_typename, "vhd", sizeof(d->d_typename));
	strlcpy(d->d_packname, "fictitious", sizeof(d->d_typename));
	d->d_secperunit = (unsigned)btodb(size);
	d->d_rpm = 7200;
	d->d_interleave = 0;
	d->d_flags = 0;
	d->d_partitions[0].p_offset = 0;
	d->d_partitions[0].p_size =
	    d->d_secperunit * (d->d_secsize / DEV_BSIZE);
	d->d_partitions[0].p_fstype = FS_BSDFFS;
	d->d_partitions[RAW_PART].p_offset = 0;
	d->d_partitions[RAW_PART].p_size =
	    d->d_secperunit * (d->d_secsize / DEV_BSIZE);
	d->d_partitions[RAW_PART].p_fstype = FS_UNUSED;
	d->d_npartitions = RAW_PART + 1;
	d->d_magic = DISKMAGIC;
	d->d_magic2 = DISKMAGIC;
	d->d_checksum = dkcksum(d);
}

/* attach a disk image */
static int
attach(const char *f, char *mnt)
{
	struct vdisk_dummy_priv	 priv;
	struct vdisk		*vd;
	vhd_header_t		 header;
	vhd_footer_t		 footer;
	uint32_t		*bat;
	pid_t			 pid;
	char			 devfile[MAXPATHLEN];
	char			 pidfile[32];
	FILE			*fp;
	int			 cc;

	(void) memset(&priv, 0x0, sizeof(priv));
	(void) memset(&header, 0x0, sizeof(header));
	(void) memset(&footer, 0x0, sizeof(footer));
	if ((fp = fopen(f, "r+")) == NULL) {
		(void) fprintf(stderr, "listvhd: can't open '%s'\n", f);
		return 0;
	}
	if (!vhd_attach(f, fp, &header, &footer, &bat)) {
		(void) fprintf(stderr, "listvhd: can't attach '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	if (geteuid() != 0) {
		(void) fprintf(stderr, "Warning: not running as root\n");
	}
	priv.size = footer.size;
	priv.buf = mmap(NULL, priv.size, PROT_READ | PROT_WRITE,
			MAP_FILE | MAP_SHARED, fileno(fp), 0);
	if (priv.buf == MAP_FAILED) {
		(void) fprintf(stderr, "can't mmap vhd image '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	if ((vd = vdisk_init(f, 0640, NULL)) == NULL) {
		(void) fprintf(stderr, "can't init vhd image '%s'\n", f);
		(void) fclose(fp);
		munmap(priv.buf, priv.size);
		return 0;
	}
	setlabel(&vd->vd_label, footer.size, header.blocksize, footer.geom);
	vd->vd_cookie = &priv;
	vd->vd_if.read = vdisk_dummy_read;
	vd->vd_if.write = vdisk_dummy_write;
	vd->vd_if.sizedev = vdisk_dummy_sizedev;
	signal(SIGINT, vdisk_signal);
	switch(pid = fork()) {
	case -1:
		(void) fprintf(stderr, "can't fork\n");
		exit(1);
	case 0:
		/* child */
		/* mount dev at mnt */
		vdisk_main(vd);
		fclose(fp);
		munmap(priv.buf, priv.size);
		vdisk_fini(vd);
		break;
	default:
		/* parent */
		cc = snprintf(devfile, sizeof(devfile), "/dev/vdisk/vd%d", vd->vd_unit);
		printf("[%d] %s %s\n", pid, devfile, mnt);
		(void) snprintf(&devfile[cc], sizeof(devfile) - cc, ".pid");
		(void) snprintf(pidfile, sizeof(pidfile), "%d", pid);
		if (symlink(pidfile, devfile) != 0) {
			(void) fprintf(stderr, "can't symlink(%s, %s)\n", pidfile, devfile);
		}
		break;
	}
	return 1;
}

/* create a disk image */
int
vdisk_create(const char *type, const char *f, const char *vhdtype, uint64_t size, uint32_t blocksize)
{
	uint64_t	 fssize;
	char		*image;

	if (strcmp(type, "vhd") == 0) {
		if ((fssize = vhd_size(vhdtype, size, blocksize)) == 0) {
			(void) fprintf(stderr, "can't create '%s'\n", f);
			return 0;
		}
		if ((image = calloc(1, fssize)) == NULL) {
			(void) fprintf(stderr, "can't create '%s'\n", f);
			return 0;
		}
		if (!vhd_create((uint8_t *)image, vhdtype, size, blocksize,
				VDISK_GEOM(256, 1, 1))) {
			(void) fprintf(stderr, "can't vhd_create '%s'\n", f);
			return 0;
		}
		return writeout(f, image, fssize);
	}
	return 0;
}

/* make a new file system, and mount it */
static int
mkfs(const char *f, char *mnt)
{
	char	mntcmd[MAXPATHLEN];

	if (!attach(f, mnt)) {
		(void) fprintf(stderr, "mkfs: can't attach %s\n", f);
		return 0;
	}
	/* XXX - hardcoded */
	if (system("newfs /dev/vdisk/rvd0a") != 0) {
		(void) fprintf(stderr, "mkfs: can't newfs %s\n", f);
		return 0;
	}
	/* XXX - hardcoded */
	(void) snprintf(mntcmd, sizeof(mntcmd), "mount /dev/vdisk/vd0a %s", mnt);
	if (system(mntcmd) != 0) {
		(void) fprintf(stderr, "mkfs: can't mount %s\n", f);
		return 0;
	}
	return 1;
}

/* unconfigure the device from the image file */
static int
unconfigure(const char *f, char *mnt)
{
	struct stat	st;
	char		pidfile[MAXPATHLEN];
	char		pid[32];
	int		cc;

	if (unmount(mnt, 0) != 0) {
		(void) fprintf(stderr, "warning: can't umount '%s'\n", mnt);
	}
	(void) snprintf(pidfile, sizeof(pidfile), "/dev/vdisk/%s.pid", f);
	if (lstat(pidfile, &st) == 0 && (st.st_mode & S_IFMT) == S_IFLNK) {
		cc = readlink(pidfile, pid, sizeof(pid));
		pid[cc] = 0x0;
		(void) kill(atoi(pid), SIGINT);
		vdisk_signal(SIGINT);
		(void) unlink(pidfile);
		return 1;
	}
	(void) fprintf(stderr, "unconfigure: %s not found\n", f);
	return 0;
}

int
vdisk_list(const char *type, const char *f)
{
	if (strcmp(type, "vhd") == 0) {
		return vhd_list(f);
	}
	return 0;
}

int
main(int argc, char **argv)
{
	const char	*vhdtype;
	uint64_t	 size;
	uint32_t	 blocksize;
	char		 mnt[MAXPATHLEN];
	char		 op;
	int		 newfs;
	int		 ex;
	int		 i;

	size = VHD_MB(64);
	blocksize = VHD_MB(2);
	op = 'a';
	vhdtype = "fixed";
	(void) snprintf(mnt, sizeof(mnt), "/mnt");
	newfs = 0;
	while ((i = getopt(argc, argv, "B:Caclm:s:t:u")) != -1) {
		switch(i) {
		case 'B':
			blocksize = atoi(optarg);
			break;
		case 'C':
			op = 'c';
			newfs = 1;
			break;
		case 'a':
		case 'c':
		case 'l':
		case 'u':
			op = (char)i;
			break;
		case 'm':
			(void) snprintf(mnt, sizeof(mnt), "%s", optarg);
			break;
		case 's':
			size = atoi(optarg);
			break;
		case 't':
			vhdtype = optarg;
			break;
		default:
			break;
		}
	}
	ex = EXIT_SUCCESS;
	switch(op) {
	case 'a':
		/* attach */
		if (!attach(argv[optind], mnt)) {
			ex = EXIT_FAILURE;
		}
		break;
	case 'c':
		/* create */
		if (!vdisk_create("vhd", argv[optind], vhdtype, size, blocksize)) {
			ex = EXIT_FAILURE;
		}
		if (newfs && !mkfs(argv[optind], mnt)) {
			ex = EXIT_FAILURE;
		}
		break;
	case 'l':
		/* list */
		vdisk_list("vhd", argv[optind]);
		break;
	case 'u':
		/* unconfigure */
		if (!unconfigure(argv[optind], mnt)) {
			ex = EXIT_FAILURE;
		}
		break;
	default:
		(void) fprintf(stderr, "unknown op '%d'\n", op);
		exit(EXIT_FAILURE);
	}
	exit(ex);
}
