/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uuid.h>

#include <inttypes.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "vmdk.h"
#include "vhd.h"
#include "virtdisk.h"

#ifndef USE_ARG
#define USE_ARG(x)  /*LINTED*/(void)&x
#endif

/* write out the image to host disk */
static int
writeout(const char *f, const char *image, uint64_t size)
{
	uint64_t	 cc;
	FILE		*fp;
	int		 wc;

	if ((fp = fopen(f, "w")) == NULL) {
		(void) fprintf(stderr, "writeout: can't open '%s'\n", f);
		return 0;
	}
	for (cc = 0 ; cc < size && (wc = write(fileno(fp), &image[cc], size - cc)) > 0 ; cc += wc) {
	}
	(void) fclose(fp);
	return cc == size;
}

/* create a disk image */
int
virtdisk_create(const char *type, const char *f, const char *vhdtype, uint64_t size, uint32_t blocksize)
{
	uint64_t	 fssize;
	char		*image;

	if (strcmp(type, "vhd") == 0) {
		if ((fssize = vhd_size(vhdtype, size, blocksize)) == 0) {
			(void) fprintf(stderr, "can't create '%s'\n", f);
			return 0;
		}
		if ((image = calloc(1, fssize)) == NULL) {
			(void) fprintf(stderr, "can't create '%s'\n", f);
			return 0;
		}
		if (!vhd_create((uint8_t *)image, vhdtype, size, blocksize,
				VDISK_GEOM(254, 16, 1))) {
			(void) fprintf(stderr, "can't vhd_create '%s'\n", f);
			return 0;
		}
		return writeout(f, image, fssize);
	} else if (strcmp(type, "vmdk") == 0) {
		if ((fssize = vmdk_size(vhdtype, size, blocksize)) == 0) {
			(void) fprintf(stderr, "can't create '%s'\n", f);
			return 0;
		}
		if ((image = calloc(1, fssize)) == NULL) {
			(void) fprintf(stderr, "can't create '%s'\n", f);
			return 0;
		}
		if (!vmdk_create((uint8_t *)image, vhdtype, size, blocksize,
				VDISK_GEOM(254, 16, 1))) {
			(void) fprintf(stderr, "can't vhd_create '%s'\n", f);
			return 0;
		}
		return writeout(f, image, fssize);
	}
	return 0;
}

/* list the device - determine the type automatically */
int
virtdisk_list(const char *f)
{
	char	type[64];

	if (!virtdisk_type(f, type, sizeof(type))) {
		return 0;
	}
	if (strcmp(type, "vhd") == 0) {
		return vhd_list(f);
	}
	if (strcmp(type, "vmdk") == 0) {
		return vmdk_list(f);
	}
	return 0;
}

/* try to determine the virtual disk type */
int
virtdisk_type(const char *f, char *type, size_t size)
{
	SparseExtentHeader	 vmdk;
	vhd_footer_t		 vhd;
	FILE			*fp;
	int			 found;

	if ((fp = fopen(f, "r")) == NULL) {
		(void) snprintf(type, size, "unknown");
		return 0;
	}
	found = 0;
	if (vmdk_read_header(fp, &vmdk) && vmdk.magicNumber == VMDK_MAGIC) {
		(void) snprintf(type, size, "vmdk");
		found = 1;
	}
	if (!found && vhd_read_footer(fp, &vhd) && vhd.version == VHD_FORMAT_VERSION) {
		(void) snprintf(type, size, "vhd");
		found = 1;
	}
	(void) fclose(fp);
	return found;
}
