/*-
 * Copyright (c) 2010 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uuid.h>

#include <inttypes.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "vmdk.h"
#include "libvdisk.h"
#include "vhd.h"

#ifndef USE_ARG
#define USE_ARG(x)  /*LINTED*/(void)&x
#endif

/* fill the header struct */
static void
fillheader(SparseExtentHeader *header, uint64_t size, int compressed, int sparse)
{
	/* we only create sparse vmdk's right now */
	(void) memset(header, 0x0, sizeof(*header));
	if (sparse) {
		header->magicNumber = VMDK_MAGIC;
		header->flags = (VALID_NEWLINE_TEST | REDUNDANT_GRAIN_TABLE);
		header->version = 1;
		header->grainSize = VMDK_DEFAULT_GRAIN_SIZE;
		header->capacity = VMDK_DEFAULT_EXTENT_SIZE;
		header->numGTEsPerGT = 32768;
		header->descriptorOffset = 1;
		header->descriptorSize = 20;
		header->rgdOffset = header->descriptorOffset + header->descriptorSize;
		header->gdOffset = header->rgdOffset + (2 * header->numGTEsPerGT) +
					(2 * header->descriptorOffset);
		header->overHead = header->gdOffset * 2;
		if (compressed) {
			header->flags |= COMPRESSED_GRAIN_TABLE;
			header->compressAlgorithm = COMPRESSION_DEFLATE;
		}
		header->singleEndLineChar = '\n';
		header->nonEndLineChar = ' ';
		header->doubleEndLineChar1 = '\r';
		header->doubleEndLineChar2 = '\n';
	} else {
		(void) fprintf(stderr, "only sparse vmdk's for now\n");
	}
}

#if 0
/* endian swapping for header struct */
static void
swapheader(SparseExtentHeader *header, const char *direction)
{
	if (strcmp(direction, "hton") == 0) {
		header->magicNumber = htobe32(header->magicNumber);
		header->version = htobe32(header->version);
		header->flags = htobe32(header->flags);
		header->capacity = htobe64(header->capacity);
		header->grainSize = htobe64(header->grainSize);
		header->descriptorOffset = htobe64(header->descriptorOffset);
		header->descriptorSize = htobe64(header->descriptorSize);
		header->numGTEsPerGT = htobe32(header->numGTEsPerGT);
		header->rgdOffset = htobe64(header->rgdOffset);
		header->gdOffset = htobe64(header->gdOffset);
		header->overHead = htobe64(header->overHead);
		header->compressAlgorithm = htobe16(header->compressAlgorithm);
	} else {
		header->magicNumber = be32toh(header->magicNumber);
		header->version = be32toh(header->version);
		header->flags = be32toh(header->flags);
		header->capacity = be64toh(header->capacity);
		header->grainSize = be64toh(header->grainSize);
		header->descriptorOffset = be64toh(header->descriptorOffset);
		header->descriptorSize = be64toh(header->descriptorSize);
		header->numGTEsPerGT = be32toh(header->numGTEsPerGT);
		header->rgdOffset = be64toh(header->rgdOffset);
		header->gdOffset = be64toh(header->gdOffset);
		header->overHead = be64toh(header->overHead);
		header->compressAlgorithm = be16toh(header->compressAlgorithm);
	}
}
#endif

/* print header info in readable format */
static int
pheader(FILE *fp, SparseExtentHeader *header)
{
	char	buf[1024];
	int	cc;

	cc = snprintf(buf, sizeof(buf), "Magic: %x\n", header->magicNumber);
	cc += snprintf(&buf[cc], sizeof(buf) - cc, "Version: %x\n",
		header->version);
	cc += snprintf(&buf[cc], sizeof(buf) - cc, "Flags: %x\n",
		header->flags);
	cc += snprintf(&buf[cc], sizeof(buf) - cc, "Capacity: %" PRIu64 "\n",
		header->capacity);
	cc += snprintf(&buf[cc], sizeof(buf) - cc, "Grain Size: %" PRIu64 "\n",
		header->grainSize);
	cc += snprintf(&buf[cc], sizeof(buf) - cc,
		"Descriptor Offset: %" PRIu64 "\n", header->descriptorOffset);
	cc += snprintf(&buf[cc], sizeof(buf) - cc,
		"Descriptor Size: %" PRIu64 "\n", header->descriptorSize);
	cc += snprintf(&buf[cc], sizeof(buf) - cc, "# GTEs / GT: %d\n",
		header->numGTEsPerGT);
	cc += snprintf(&buf[cc], sizeof(buf) - cc,
		"Redundant Metadata level 0: %" PRIu64 "\n", header->rgdOffset);
	cc += snprintf(&buf[cc], sizeof(buf) - cc,
		"Metadata sectors (overhead): %" PRIu64 "\n", header->overHead);
	cc += snprintf(&buf[cc], sizeof(buf) - cc, "Unclean shutdown: %s\n",
		(header->uncleanShutdown) ? "true" : "false");
	cc += snprintf(&buf[cc], sizeof(buf) - cc, "Compression algorithm: %d\n",
		header->compressAlgorithm);
	return fwrite(buf, 1, cc, fp) == (size_t)cc;
}

/* dump vmdk disk info */
static int
fmtdiskinfo(const char *image, uint64_t bytes, char *buf, size_t size)
{
	int	cc;

	cc = 0;
	cc += snprintf(&buf[cc], size - cc, "# Disk DescriptorFile\n");
	cc += snprintf(&buf[cc], size - cc, "version=1\n");
	cc += snprintf(&buf[cc], size - cc, "CID=%x\n", (unsigned)random());
	cc += snprintf(&buf[cc], size - cc, "parentCID=ffffffff\n");
	cc += snprintf(&buf[cc], size - cc, "createType=\"monolithicSparse\"\n\n");
	cc += snprintf(&buf[cc], size - cc, "# Extent description\n");
	cc += snprintf(&buf[cc], size - cc,
		"RW %" PRIu64 " SPARSE \"%s\"\n\n", bytes / VMDK_SECTOR_SIZE, image);
	cc += snprintf(&buf[cc], size - cc, "# The Disk Data Base\n");
	cc += snprintf(&buf[cc], size - cc, "#DDB\n\n");
	cc += snprintf(&buf[cc], size - cc, "ddb.adapterType = \"ide\"\n");
	cc += snprintf(&buf[cc], size - cc, "ddb.geometry.sectors = \"63\"\n");
	cc += snprintf(&buf[cc], size - cc, "ddb.geometry.heads = \"16\"\n");
	cc += snprintf(&buf[cc], size - cc, "ddb.geometry.cylinders = \"16383\"\n");
	cc += snprintf(&buf[cc], size - cc, "ddb.virtualHWVersion = \"6\"\n");
	cc += snprintf(&buf[cc], size - cc, "ddb.toolsVersion = \"0\"\n");
	return 1;
}

/***************************************************************************/

/* create a vmdk header, footer and BAT */
int
vmdk_create(uint8_t *image, const char *vhdtype, uint64_t size, uint32_t blocksize, uint32_t geom)
{
	SparseExtentHeader	header;
	char			diskinfo[1024];

	(void) memset(&header, 0x0, sizeof(header));
	fillheader(&header, size, 0, 1);
#if 0
	swapheader(&header, "hton");
#endif
	(void) memcpy(image, &header, sizeof(header));
	(void) memset(diskinfo, 0x0, sizeof(diskinfo));
	fmtdiskinfo("dummy", size, diskinfo, sizeof(diskinfo));
	(void) memcpy(&image[1024], diskinfo, sizeof(diskinfo));
	return 1;
}

/* XXX */
uint64_t
vmdk_size(const char *type, uint64_t size, uint32_t blocksize)
{
	USE_ARG(type);
	USE_ARG(blocksize);
	return size;
}

/* read the sparse extent header */
int
vmdk_read_header(FILE *fp, SparseExtentHeader *header)
{
	if (read(fileno(fp), header, sizeof(*header)) != sizeof(*header)) {
		return 0;
	}
#if 0
	swapheader(header, "ntoh");
#endif
	return 1;
}

/* read a vmdk header and ddb */
int
vmdk_list(const char *f)
{
	SparseExtentHeader	 header;
	char			 diskinfo[1024];
	FILE			*fp;

	(void) memset(&header, 0x0, sizeof(header));
	if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "vmdk_list: can't open '%s'\n", f);
		return 0;
	}
	if (!vmdk_read_header(fp, &header)) {
		(void) fprintf(stderr, "vmdk_list: bad read '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	printf("[%s]\n", f);
	pheader(stdout, &header);
	if (pread(fileno(fp), diskinfo, sizeof(diskinfo), 1024) != sizeof(diskinfo)) {
		(void) fprintf(stderr, "vmdk_list: bad read '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	printf("\n%s\n", diskinfo);
	(void) fclose(fp);
	return 1;
}
